// Copyright 2022 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
import { createInterface } from 'node:readline';
export async function question(query, { choices } = {}) {
    let completer = undefined;
    if (Array.isArray(choices)) {
        completer = function completer(line) {
            const completions = choices;
            const hits = completions.filter((c) => c.startsWith(line));
            return [hits.length ? hits : completions, line];
        };
    }
    const rl = createInterface({
        input: process.stdin,
        output: process.stdout,
        terminal: true,
        completer,
    });
    return new Promise((resolve) => rl.question(query ?? '', (answer) => {
        rl.close();
        resolve(answer);
    }));
}
